#! /bin/bash

#**************************************************************************
#    begin                : Wed Feb 28, 2001
#    copyright            : (C) 2001-2004 by Adaptec
#    email                : mark_salyzyn@adaptec.com
#    original author      : deanna_bonds@adaptec.com
#
#    This is the installer script for Adaptec's RAID drivers
#**************************************************************************
#                                                                         *
#   Portions of this program are free software; you can redistribute it   *
#   and/or modify it under the terms of the GNU General Public License    *
#   as published by the Free Software Foundation; either version 2 of     *
#   the License, or (at your option) any later version.                   *
#                                                                         *
#**************************************************************************

LANG=C
LC_CTYPE=C
export LANG LC_CTYPE

product=""
branch=""
installer_path=""

case `pwd` in				# In directory
*aar1210*)	product=aar1210 ; branch=`pwd|sed 's#.*aar1210\([^/]*\).*#\1#'` ;;
*a320raid*)	product=a320raid ; branch=`pwd|sed 's#.*a320raid\([^/]*\).*#\1#'` ;;
*aacraid*)	product=aacraid ; branch=`pwd|sed 's#.*aacraid\([^/]*\).*#\1#'` ;;
*dpt_i2o*)	product=dpt_i2o ; branch=`pwd|sed 's#.*dpt_i2o\([^/]*\).*#\1#'` ;;
*)	case $0 in			# In script path
*aar1210*)	product=aar1210 ; branch=`echo $0|sed 's#.*aar1210\([^/]*\).*#\1#'` ;;
*a320raid*)	product=a320raid ; branch=`echo $0|sed 's#.*a320raid\([^/]*\).*#\1#'` ;;
*aacraid*)	product=aacraid ; branch=`echo $0|sed 's#.*aacraid\([^/]*\).*#\1#'` ;;
*dpt_i2o*)	product=dpt_i2o ; branch=`echo $0|sed 's#.*dpt_i2o\([^/]*\).*#\1#'` ;;
*)	case $1 in			# In argument
aar1210)	product=aar1210 ;;
a320raid)	product=a320raid ;;
aacraid)	product=aacraid ;;
dpt_i2o)	product=dpt_i2o ;;
*)		if [ -z "${product}" ] ; then
			echo
			echo "****** PROBLEM *******"
			echo "Unknown product"
			echo "**********************"
			exit 1
		fi ;;
		esac ;;
	esac
	branch=$2 ;;
esac
if [ -z "${installer_path}" ] ; then
	installer_path=/opt/Adaptec/${product}${branch}
fi
PATH=${installer_path}:${PATH}
export PATH

VERSION="1.5"
verbose=0
root_dir=""
do_mkinitrd=1
do_lilo=1
force=0
module_count=0
use_grub=0
echo
echo Adaptec ${product} driver installer - V${VERSION}
echo

lilo_conf=${root_dir}"/etc/lilo.conf"
grub_conf=${root_dir}"/boot/grub/grub.conf"
if ! [ -e ${grub_conf} ] ; then
	grub_conf=${root_dir}"/boot/grub/menu.lst"
fi
if [ -s ${installer_path}/${product}"_prebuilt.tgz" ] ; then
	prebuilt_binaries=${installer_path}/${product}"_prebuilt.tgz"
elif [ -s ${installer_path}/modules.cgz ] ; then
	prebuilt_binaries=${installer_path}/modules.cgz
elif [ -d ${installer_path}/linux ] ; then
	prebuilt_binaries=${installer_path}/linux
elif [ -d ${installer_path}/objects ] ; then
	prebuilt_binaries=${installer_path}/objects
else
	echo
	echo "****** PROBLEM *******"
	echo "Unknown source of prebuilt binaries"
	echo "**********************"
	return 1
fi
tmp_magic=/tmp/magic$$

# output usage statement
usage () {
	echo "usage: `basename $0` [--version] || [--help] ||" >&2
	echo "            [--verbose] [--root=<root dir>] [--noinitrd] [--nolilo] [--force]" >&2
	echo
	echo "usage: `basename $0` [--list_modules] " >&2
	echo
	echo "For the proceedure to patch your kernel sources see ${installer_path}/Read.me"
	echo
	exit 1
}

# write an updated magic file
# for those illiterate in the magic format, here is
# some `C' code to help you interpret the following:
#
#	if (memcmp((char *)0x202, "HdrS", 4) == 0) {
#		printf (" Linux/i386 Kernel");
#		if (*((unsigned short *)0x206) == 0x105) {
#			printf (" setupV1 zImage");
#		} else if (*((unsigned short *)0x206) == 0x200) {
#			printf (" setupV2");
#			if (*((unsigned char *)0x211) == 0) {
#				printf (" zImage");
#			} else if (*((unsigned char *)0x211) == 1) {
#				printf (" bzImage");
#			}
#		} else if (*((char *)(0x200+5+*((unsigned short *)0x20E))) == ' ') {
#			printf (" %5.5s", 0x200+*((unsigned short *)0x20E));
#		} else if (*((char *)(0x200+6+*((unsigned short *)0x20E))) == ' ') {
#			printf (" %6.6s", 0x200+*((unsigned short *)0x20E));
#		} else if (*((char *)(0x200+7+*((unsigned short *)0x20E))) == ' ') {
#			printf (" %7.7s", 0x200+*((unsigned short *)0x20E));
#		. . .
#	}
#
write_magic () {
cat <<-blub >${tmp_magic}
	0x202	string		HdrS		Linux/i386 Kernel
	>0x206	short		0x0105		setupV1 zImage
	>0x206	short		0x0200		setupV2
	>>0x211	byte		0		zImage
	>>0x211	byte		1		bzImage
	>(0x20e.s+0x205)	byte	32
	>>(0x20e.s+0x200)	string	>\0	"%5.5s"
	>(0x20e.s+0x206)	byte	32
	>>(0x20e.s+0x200)	string	>\0	"%6.6s"
	>>>(0x20e.s+0x21F)	string	Mon\ Mar\ 31\ 23:11:24\ UTC\ 2003	"SuSE-x86_64-189"
	>>>(0x20e.s+0x21F)	string	Mon\ Mar\ 31\ 23:47:50\ UTC\ 2003	"SuSE-x86_64-191"
	>(0x20e.s+0x207)	byte	32
	>>(0x20e.s+0x200)	string	>\0	"%7.7s"
	>(0x20e.s+0x208)	byte	32
	>>(0x20e.s+0x200)	string	>\0	"%8.8s"
	>(0x20e.s+0x209)	byte	32
	>>(0x20e.s+0x200)	string	>\0	"%9.9s"
	>(0x20e.s+0x20a)	byte	32
	>>(0x20e.s+0x200)	string	>\0	"%10.10s"
	>>>(0x20e.s+0x224)	string	Mon\ Oct\ 21\ 18:39:02\ UTC\ 2002	"SuSE-athlon-111"
	>>>(0x20e.s+0x224)	string	Thu\ Mar\ 20\ 14:35:44\ UTC\ 2003	"SuSE-athlon-263"
	>>>(0x20e.s+0x224)	string	Wed\ Apr\ 16\ 12:40:55\ UTC\ 2003	"SuSE-athlon-290"
	>>>(0x20e.s+0x224)	string	Tue\ May\ 20\ 08:20:18\ UTC\ 2003	"SuSE-athlon-304"
	>>>(0x20e.s+0x224)	string	Mon\ Aug\ 4\ 23:24:04\ UTC\ 2003	"SuSE-athlon-340"
	>>>(0x20e.s+0x224)	string	Tue\ Sep\ 30\ 19:01:07\ UTC\ 2003	"SuSE-athlon-343"
	>>>(0x20e.s+0x224)	string	Mon\ Oct\ 21\ 18:28:11\ UTC\ 2002	"SuSE-i686-82"
	>>>(0x20e.s+0x225)	string	Mon\ Oct\ 21\ 18:45:41\ UTC\ 2002	"SuSE-i686-120"
	>>>(0x20e.s+0x225)	string	Thu\ Mar\ 20\ 14:37:45\ UTC\ 2003	"SuSE-i686-274"
	>>>(0x20e.s+0x225)	string	Wed\ Apr\ 16\ 17:09:34\ UTC\ 2003	"SuSE-i686-290"
	>>>(0x20e.s+0x225)	string	Tue\ May\ 20\ 08:20:00\ UTC\ 2003	"SuSE-i686-304"
	>>>(0x20e.s+0x225)	string	Wed\ May\ 21\ 18:12:39\ UTC\ 2003	"SuSE-i686-306"
	>>>(0x20e.s+0x225)	string	Mon\ Aug\ 4\ 23:23:04\ UTC\ 2003	"SuSE-i686-340"
	>>>(0x20e.s+0x225)	string	Tue\ Sep\ 30\ 19:03:19\ UTC\ 2003	"SuSE-i686-343"
	>>>(0x20e.s+0x227)	string	Mon\ Mar\ 31\ 23:48:08\ UTC\ 2003	"SuSE-x86_64-191"
	>(0x20e.s+0x20b)	byte	32
	>>(0x20e.s+0x200)	string	>\0	"%11.11s"
	>>>(0x20e.s+0x228)	string	Mon\ Mar\ 31\ 23:49:47\ UTC\ 2003	"SuSE-x86_64-201"
	>(0x20e.s+0x20c)	byte	32
	>>(0x20e.s+0x200)	string	>\0	"%12.12s"
	>(0x20e.s+0x20d)	byte	32
	>>(0x20e.s+0x200)	string	>\0	"%13.13s"
	>(0x20e.s+0x20e)	byte	32
	>>(0x20e.s+0x200)	string	>\0	"%14.14s"
	>>>(0x20e.s+0x22d)	string	Wed\ Apr\ 16\ 12:40:22\ UTC\ 2003	"SuSE-i686-290"
	>>>(0x20e.s+0x22d)	string	Thu\ Mar\ 20\ 14:36:54\ UTC\ 2003	"SuSE-i686-263"
	>>>(0x20e.s+0x22d)	string	Mon\ Oct\ 21\ 18:44:17\ UTC\ 2002	"SuSE-i686-115"
	>>>(0x20e.s+0x22d)	string	Tue\ May\ 20\ 08:19:43\ UTC\ 2003	"SuSE-i686-304"
	>>>(0x20e.s+0x22d)	string	Wed\ May\ 21\ 18:12:23\ UTC\ 2003	"SuSE-i686-306"
	>>>(0x20e.s+0x22d)	string	Mon\ Aug\ 4\ 23:25:58\ UTC\ 2003	"SuSE-i686-340"
	>>>(0x20e.s+0x22d)	string	Tue\ Sep\ 30\ 19:02:16\ UTC\ 2003	"SuSE-i686-343"
	>(0x20e.s+0x20f)	byte	32
	>>(0x20e.s+0x200)	string	>\0	"%15.15s"
	>>>(0x20e.s+0x22e)	string	Wed\ Mar\ 27\ 13:58:12\ UTC\ 2002	"SuSE-i686-35"
	>>>(0x20e.s+0x22e)	string	Mon\ Oct\ 21\ 18:48:05\ UTC\ 2002	"SuSE-i686-113"
	>>>(0x20e.s+0x230)	string	Thu\ Mar\ 20\ 14:44:44\ UTC\ 2003	"SuSE-debug-213"
	>>>(0x20e.s+0x22e)	string	Thu\ Mar\ 20\ 14:41:55\ UTC\ 2003	"SuSE-i686-257"
	>>>(0x20e.s+0x230)	string	Wed\ Apr\ 16\ 12:42:04\ UTC\ 2003	"SuSE-debug-290"
	>>>(0x20e.s+0x22e)	string	Wed\ Apr\ 16\ 12:42:15\ UTC\ 2003	"SuSE-i686-290"
	>>>(0x20e.s+0x230)	string	Tue\ May\ 20\ 08:19:29\ UTC\ 2003	"SuSE-debug-304"
	>>>(0x20e.s+0x22e)	string	Tue\ May\ 20\ 08:20:31\ UTC\ 2003	"SuSE-i686-304"
	>>>(0x20e.s+0x22e)	string	Wed\ May\ 21\ 18:11:18\ UTC\ 2003	"SuSE-i686-306"
	>>>(0x20e.s+0x230)	string	Mon\ Aug\ 4\ 23:27:19\ UTC\ 2003	"SuSE-debug-340"
	>>>(0x20e.s+0x22e)	string	Mon\ Aug\ 4\ 23:48:22\ UTC\ 2003	"SuSE-i686-340"
	>>>(0x20e.s+0x22e)	string	Wed\ Oct\ 1\ 10:12:46\ UTC\ 2003	"SuSE-i686-343"
	>(0x20e.s+0x210)	byte	32
	>>(0x20e.s+0x200)	string	>\0	"%16.16s"
	>(0x20e.s+0x211)	byte	32
	>>(0x20e.s+0x200)	string	>\0	"%17.17s"
	>>>(0x20e.s+0x22b)	string	Wed\ May\ 21\ 18:12:07\ UTC\ 2003	"SuSE-athlon-306"
	>(0x20e.s+0x212)	byte	32
	>>(0x20e.s+0x200)	string	>\0	"%18.18s"
	>(0x20e.s+0x213)	byte	32
	>>(0x20e.s+0x200)	string	>\0	"%19.19s"
	>(0x20e.s+0x214)	byte	32
	>>(0x20e.s+0x200)	string	>\0	"%20.20s"
	>(0x20e.s+0x215)	byte	32
	>>(0x20e.s+0x200)	string	>\0	"%21.21s"
	>(0x20e.s+0x216)	byte	32
	>>(0x20e.s+0x200)	string	>\0	"%22.22s"
	>(0x20e.s+0x217)	byte	32
	>>(0x20e.s+0x200)	string	>\0	"%23.23s"
	>(0x20e.s+0x218)	byte	32
	>>(0x20e.s+0x200)	string	>\0	"%24.24s"
	>(0x20e.s+0x219)	byte	32
	>>(0x20e.s+0x200)	string	>\0	"%25.25s"
	>(0x20e.s+0x21a)	byte	32
	>>(0x20e.s+0x200)	string	>\0	"%26.26s"
	>(0x20e.s+0x21b)	byte	32
	>>(0x20e.s+0x200)	string	>\0	"%27.27s"
	>(0x20e.s+0x21c)	byte	32
	>>(0x20e.s+0x200)	string	>\0	"%28.28s"
	>(0x20e.s+0x21d)	byte	32
	>>(0x20e.s+0x200)	string	>\0	"%29.29s"
	>(0x20e.s+0x21e)	byte	32
	>>(0x20e.s+0x200)	string	>\0	"%30.30s"
blub
}

clean_magic () {
	rm -f ${tmp_magic} 2>/dev/null
}

list_binary_equivalent_modules () {
	if [ -x /usr/bin/perl ] ; then
		sed -e 's/[ 	]*#.*//' -e 's/[ 	]*$//' -e '/^\$/d' ${installer_path}/module.equiv
	fi
}

list_prebuilt_modules () {

	echo Prebuilt modules:
	case ${prebuilt_binaries} in
	*.tgz|*.tar.gz)
		list=`tar ztf ${prebuilt_binaries} 2>/dev/null | sed "s/${product}-//g" | sort -u` ;;
	*.cgz|.cpio.gz)
		list=`gunzip <${prebuilt_binaries} | cpio -it 2>/dev/null | sed "s#/${product}.k*o##g" | sort -u` ;;
	*/linux)
		list=`for i in ${prebuilt_binaries}/*/*/install/update.tar.gz ; do arch="\`echo $i | sed -n 's#.*/\([^/]*\)/install.*#/\1/p'\`" ; tar ztf ${i} 2>/dev/null | sed "s/$/ ${arch}/" ; done | sed -n "s#.*lib/modules/\([^/]*\)/.*/${product}.k*o#\1#p" | sort -u` ;;
	*/objects)
		list=`if [ -d ${prebuilt_binaries}/${product} ] ; then ls ${prebuilt_binaries}/${product}/ ; else ls ${prebuilt_binaries}/ ; fi | sed -n "s#${product}-##p" | sort -u` ;;
	esac
	for i in ${list} ; do
		echo "	** ${i}"
		# Any equivalents that are fulfilled?
		list_binary_equivalent_modules 2>/dev/null | while read orig equiv ; do
			if [ "${product}-${i}" == "${orig}" ] ; then
				# do we have it anyways?
				ok=yes
				for j in ${list} ; do
					if [ "${product}-${j}" == "${equiv}" ] ;then
						ok=no
						break
					fi
				done
				if [ "${ok}" == "yes" ] ; then
					echo "	** +${equiv}"
				fi
			fi
		done | sed -e "s/${product}-//" -e '$s/\*\* ./\*\* \\/'
	done

	echo
}

extract_the_prebuilt_modules () {

	case ${prebuilt_binaries} in
	*.tgz|*.tar.gz)
		module_file=${product}-${1}${2}${3}
		tar zxf ${prebuilt_binaries} ${module_file} >/dev/null 2>/dev/null ;;
	*.cgz|.cpio.gz)
		module_file=${1}${2}${3}/${product}.o
		gunzip <${prebuilt_binaries} | cpio -imdu ${module_file} >/dev/null 2>/dev/null
		if [ -s ${module_file} ] ; then
			break
		fi
		module_file=${1}${3}/`echo ${2} | sed 's/-//'`/${product}.o
		gunzip <${prebuilt_binaries} | cpio -imdu ${module_file} >/dev/null 2>/dev/null
		if [ -s ${module_file} ] ; then
			break
		fi
		module_file=${1}${3}/${product}.o
		gunzip <${prebuilt_binaries} | cpio -imdu ${module_file} >/dev/null 2>/dev/null ;;
	*/linux)
		for i in ${prebuilt_binaries}/*/`echo ${2} | sed 's/-\(.*\)/\1-/'`*/install/update.tar.gz ; do
			tar zxf ${i} >/dev/null 2>/dev/null
			module_file=lib/modules/${1}${3}/kernel/drivers/scsi/${product}.o
			if [ -s ${module_file} ] ; then
				break
			fi
			module_file=lib/modules/${1}${3}/kernel/drivers/scsi/${product}.ko
			if [ -s ${module_file} ] ; then
				break
			fi
			module_file=lib/modules/${1}${3}/kernel/drivers/scsi/${product}/${product}.o
			if [ -s ${module_file} ] ; then
				break
			fi
			module_file=lib/modules/${1}${3}/kernel/drivers/scsi/${product}/${product}.ko
			if [ -s ${module_file} ] ; then
				break
			fi
		done ;;
	*/objects)
		if [ -s ${prebuilt_binaries}/${product}/${product}-${1}${2}${3} ] ; then
			module_file=${prebuilt_binaries}/${product}/${product}-${1}${2}${3}
		else
			module_file=${prebuilt_binaries}/${product}-${1}${2}${3}
		fi ;;
	esac
}

extract_prebuilt_modules () {
	extract_the_prebuilt_modules ${1} ${2} ${3}
	if [ ! -e ${module_file} ] ; then
		temp=/tmp/install.sh.$$
		trap 'rm -f ${temp} ; exit' 1 2 3 15
		rm -f ${temp}
		list_binary_equivalent_modules 2>/dev/null | while read orig equiv ; do
			if [ "${product}-${1}${2}${3}" == "${equiv}" ] ; then
				orig_kern=`echo ${orig} | sed "s/${product}-//"`
				extract_the_prebuilt_modules ${orig_kern}
				if [ -e ${module_file} ] ; then
					equiv_kern=`echo ${equiv} | sed "s/${product}-//"`
					for suffix in -i386 -i486 -i586 -i686 -athlon -64GB -ia32e -ia32 -ia64 -x86_64 -64GB -enterprise -debug -smp4G -psmp -smp -summit -bigmem -uml -hugemem -SMP -default -itanium2 -page-64k enterprise debug smp4G psmp smp summit bigmem uml hugemem SMP default itanium2 page-64k ; do
						equiv_kern=`echo ${equiv_kern} | sed "s/${suffix}//"`
						orig_kern=`echo ${orig_kern} | sed "s/${suffix}//"`
					done
					# Binary must remain same size ...
					if [ `echo ${equiv_kern} | sed 's/./@/g'` == `echo ${orig_kern} | sed 's/./@/g'` ] ; then
						# struct_module also needs to
						# be adjusted ...
						/usr/bin/perl -e 'while (<>) {
							s/'${orig_kern}/${equiv_kern}'/g;
							print;
						}' <${module_file} >`basename ${module_file}-${equiv_kern}`
						case ${module_file} in
						*/objects/*${product}-*) ;;
						*) rm -f ${module_file} ;;
						esac
						module_file=`basename ${module_file}-${equiv_kern}`
						echo ${module_file} > ${temp}
						break
					else
						module_file=
						echo
						echo "****** PROBLEM *******"
						echo "Illegal entry in module.equiv (kernel names are not the same length):"
						echo "	"${orig} ${equiv}
						echo "**********************"
					fi
				fi
			fi
		done
		if [ -s ${temp} ] ; then
			module_file=`cat ${temp}`
		fi
		rm -f ${temp}
		trap 'exit' 1 2 3 15
	fi
}

module_failed_before() {
	if [ ! -s /tmp/${product}.fail ] ; then
		return 1
	fi
	while read one ; do
		if [ "${one}" == "$1" ] ; then
			return 0
		fi
	done < /tmp/${product}.fail
	return 1
}

install_module() {
	if ! [ -e $1 ] ; then
		mkdir -p $1
	fi
	cd $1

	versions="ia64 ia32e x86_64 athlon i686 i586 i486 i386 ia32"

	if [ ${verbose} -eq 1 ] ; then
		echo Installing Module...
	fi

	firstloop=1
	for m in ${versions} ; do
		if [ ${firstloop} -eq 1 ] ; then
			if [ ${m} == $3 ] ; then
				firstloop=0
			else
				continue
			fi
		fi
		if [ ${m} == "" ] ; then
			m="ia32"
		fi
		m="-${m}"
		case $2 in
		*${m})	m="" ;;
		esac
		if [ ${verbose} -eq 1 ] ; then
			echo "Checking for prebuilt $2${m}."
		fi
		build="`echo $4 | sed -n 's/.*-.*\(-[0-9]*\)\$/\1/p'`"
		if [ ! -z "${build}" ] ; then
			extract_prebuilt_modules $2 ${m} ${build}
			if [ -e ${module_file} ] ; then
				if [ ${verbose} -eq 1 ] ; then
					echo "Found prebuilt for $2${m}."
				fi
				break;
			fi
		fi
		extract_prebuilt_modules $2 ${m}
		if [ -e ${module_file} ] ; then
			if [ ${verbose} -eq 1 ] ; then
				echo "Found prebuilt for $2${m}."
			fi
			break;
		fi
	done
	if ! [ -e "${module_file}" ] ; then
		case ${prebuilt_binaries} in
		*/linux)	rm -rf lib ;;
		esac
		if ! module_failed_before $2 || [ ${verbose} -eq 1 ] ; then
			echo
			echo "****** PROBLEM *******"
			echo "No prebuilt ${product} module for "$2 $3
			echo "**********************"
		fi
		return 1
	fi
	depend="`echo \"$1\" | sed -e 's#/kernel/drivers/scsi##' -e 's#/scsi##'`"/modules.dep
	if [ -d "${product}" ] ; then
		dest=${product}/${product}.o
		if [ -s ${product}/${product}.ko ] ; then
			dest=${product}/${product}.ko
		elif [ ! -s ${product}/${product}.o ] ; then
			case ${depend} in
			/lib/modules/2.6.*)
				dest=${product}/${product}.ko ;;
			esac
		fi
		if [ -s "`basename ${dest}`" ] ; then
			rm -f "`basename ${dest}`"
		fi
		if [ -s "${depend}" ] ; then
			if grep "scsi/${dest}" "${depend}" >/dev/null 2>/dev/null ; then
				sed "#scsi/${product}\.k*o#d" "${depend}"
			else
				sed "s#scsi/${product}\.k*o#scsi/${dest}#" "${depend}"
			fi >"${depend}.$$"
		fi
	else
		dest=${product}.o
		if [ -s ${product}.ko ] ; then
			dest=${product}.ko
		elif [ ! -s ${product}.o ] ; then
			case ${depend} in
			/lib/modules/2.6.*)
				dest=${product}.ko ;;
			esac
		fi
		if [ -s "${depend}" ] ; then
			if grep "scsi/${dest}" "${depend}" >/dev/null 2>/dev/null ; then
				sed "#scsi/${product}/${product}\.k*o#d" "${depend}"
			else
				sed "s#scsi/${product}/${product}\.k*o#scsi/${dest}#" "${depend}"
			fi >"${depend}.$$"
		fi
	fi
	if ! grep "scsi/${dest}" "${depend}.$$" >/dev/null 2>/dev/null ; then
		echo "`echo ${depend} | sed 's#modules.dep$##'`kernel/drivers/scsi/${dest}: `echo ${depend} | sed 's#modules.dep$##'`kernel/drivers/scsi/scsi_mod`echo ${dest} | sed 's#.*\(\.k*o\)#\1#'`" >> "${depend}.$$"
	fi
	case ${module_file} in
	*/objects/*${product}-*)
		cp ${module_file} ${dest} ;;
	*)
		mv -f ${module_file} ${dest}
		rmdir -p ${module_file} >/dev/null 2>/dev/null ;;
	esac
	case ${prebuilt_binaries} in
	*/linux)	rm -rf lib ;;
	esac
	if [ -s "${depend}.$$" ] && ! cmp "${depend}" "${depend}.$$" >/dev/null 2>/dev/null ; then
		if [ -e "${depend}" ] ; then
			mv -f "${depend}" "${depend}.orig"
		fi
		mv "${depend}.$$" "${depend}"
	else
		rm -f "${depend}.$$"
	fi
	# clean up anything that was left over from an install or other
	# activity that would block this driver from being incorporated into
	# an initrd.
	rm -f "`echo ${depend} | sed 's#modules.dep$##'`updates/"${product}.[ko]*
	if [ ${verbose} -eq 1 ] ; then
		echo "Module installed for "$1" kernel."
	fi
	return 0
}


modify_rc_config() {
	rc_config=${root_dir}"/etc/sysconfig/kernel"

	if ! [ -e ${rc_config} ] ; then
		rc_config=${root_dir}"/etc/rc.config"
		if ! [ -e ${rc_config} ] ; then
			if [ ${verbose} -eq 1 ] ; then
				echo "No rc.config file"
			fi
			return 1
		fi
	fi

	if [ ${verbose} -eq 1 ] ; then
		echo "Modifying "${rc_config}
	fi

        # Backup current config
        cp -fp ${rc_config} ${rc_config}.$$
	if [ ${verbose} -eq 1 ] ; then
		grep "INITRD_MODULES" ${rc_config}
	else
		grep "INITRD_MODULES" ${rc_config} >/dev/null
	fi
        if [ $? -ne 0 ] ; then
		# Configuration not set up to use ramdisks add it
		echo "INITRD_MODULES= \"\"" >>${rc_config}
        fi
	sed "s/INITRD_MODULES[ \t]*=[ \t]*/INITRD_MODULES=/" ${rc_config} >/tmp/rc.config.tmp && mv -f /tmp/rc.config.tmp ${rc_config}

        current_modules=`grep "INITRD_MODULES[ \t]*=" ${rc_config} | head -1`
        echo ${current_modules} | grep ${product} >/dev/null
	if [ $? -ne 0 ] ; then
		# If we cannot find in the config file, add it.
		new_modules=`echo ${current_modules} | sed "s/\"[ \t]*$/ ${product}\"/"`
		sed "s/INITRD_MODULES.*/${new_modules}/" ${rc_config} >/tmp/rc.config.tmp && mv -f /tmp/rc.config.tmp ${rc_config}
        fi
	if [ ${verbose} -eq 1 ] ; then
		grep "INITRD_MODULES" ${rc_config}
	fi
	return 0
}


modify_modules_conf() {

	modules_conf=""
	if [ -e ${root_dir}"/etc/modules.conf" ] ; then
		modules_conf="${root_dir}/etc/modules.conf"
	elif [ -e ${root_dir}"/etc/conf.modules" ] ; then
		modules_conf=${root_dir}"/etc/conf.modules"
	else
		if [ ${verbose} -eq 1 ] ; then
			echo "No modules.conf file"
		fi
		return 1
	fi
	if [ ${verbose} -eq 1 ] ; then
		echo "Modifying "${modules_conf}
	fi

	grep ${product} ${modules_conf} > /dev/null
	if [ $? -ne 0 ] ;  then
		# Put in entry if not already in config file
		echo "alias scsi_hostadapter ${product}" >> ${modules_conf}
		if [ ${verbose} -eq 1 ] ; then
			echo "No ${product} entry - adding it..."
		fi
	fi

	tmp_modules=/tmp/modules.conf.$$
	# Enumerate scsi_hostadapter entries
	awk ' \
	BEGIN \
	{ \
		c=1; \
	} \
	{ \
		if (($1 == "alias") && (substr($2, 0, length("scsi_hostadapter")) == "scsi_hostadapter")) \
		{ \
			$2="scsi_hostadapter"c; \
			print $0; \
			c=c+1; \
		} \
		else \
		{ \
			print $0; \
		} \
	}' ${modules_conf} >${tmp_modules} && mv -f ${tmp_modules}  ${modules_conf}
	if [ ${verbose} -eq 1 ] ; then
		grep ${product} ${modules_conf}
	fi

}


# Making the initrd is one of the things that are very different
# between the different distributions.  This script has been tested
# with RedHat, SuSE, Mandrake, and Caldera

run_mkinitrd() {

        echo "Creating initrd..."
# $1 = image
# $2 = kernel version
# $3 = initrd

	old_initrd=$3.$$
	mv -f $3 ${old_initrd} 2>/dev/null >/dev/null
	mkinitrd=`which mk_initrd 2>/dev/null`

	if [ -z "${mkinitrd}" ] ; then
		# if not the SuSE mk_initrd then try Red Hat/Mandrake style
		if [ "${root_dir}" != "" ] ; then
			# if we specified root then use our modified one
			mkinitrd="${installer_path}/adpt_mkinitrd --root="${root_dir}
		else
			mkinitrd=`which mkinitrd 2>/dev/null`
		fi

		if [ -z "${mkinitrd}" ] ; then
			# not Red Hat - check if Caldera
			tmp_initrd=`echo $3 | sed s/.gz$//`
			if [ "${tmp_initrd}" != "${initrd}" ] ; then
				if [ ${verbose} -eq 1 ] ; then
					echo "Making Caldera initrd..."
				fi
				if ! [ -e /initrd ] ; then
					mkdir -f /initrd
				fi

				mv -f ${old_initrd} $3

				rootfs="/etc/modules/rootfs"
				grep scsi_mod ${rootfs} >/dev/null 2>/dev/null
				if [ $? -ne 0 ] ; then
					echo scsi_mod >>${rootfs}
				fi
				grep ${product} ${rootfs} >/dev/null 2>/dev/null
				if [ $? -ne 0 ] ; then
					echo ${product} >>${rootfs}
				fi
				grep sd ${rootfs} >/dev/null 2>/dev/null
				if [ $? -ne 0 ] ; then
					echo sd >>${rootfs}
				fi
				gunzip $3 >/dev/null
				mount -o loop ${tmp_initrd} /initrd >/dev/null 2>/dev/null
				if [ -s /lib/modules/$2/kernel/drivers/scsi/${product}/${product}.o ] ; then
					cp -f /lib/modules/$2/kernel/drivers/scsi/${product}/${product}.o /initrd/lib/modules/
				else
					cp -f /lib/modules/$2/kernel/drivers/scsi/${product}.o /initrd/lib/modules/
				fi
				cp -f ${rootfs} /initrd/etc/modules/rootfs
				sync
				umount ${tmp_initrd} >/dev/null 2>/dev/null
				sync
				gzip -9 ${tmp_initrd} >/dev/null
				depmod -qa $2
				return 0
			else
				echo "Can not find mkinitrd. Exitting..."
				exit 1
			fi
		fi

		if [ ${verbose} -eq 1 ] ; then
			echo "${mkinitrd} $3 $2"
		fi

		${mkinitrd} $3 $2
	else
		grep root_device ${mkinitrd} >/dev/null 2>/dev/null
		if [ $? -ne 0 ] ; then
			mkinitrd="${installer_path}/adpt_mk_initrd"
		fi

		if [ "${root_dir}" != "" ] ; then
			mkinitrd=${mkinitrd}" "${root_dir}
		fi
		if [ ${verbose} -eq 1 ] ; then
			echo "${mkinitrd} -k `basename $1` -i `basename $3`"
		fi

		${mkinitrd}  -k `basename $1` -i `basename $3`
	fi

	if [ $? -ne 0 -o ! -s $3 ] ; then
		echo Problem making initrd.  Quitting.
		mv -f ${old_initrd} $3
		return 1
	fi
	mv -f ${old_initrd} $3.orig

	# Check size for any clues of failures.
	orig_size=`wc -c $3.orig | sed -n 's/[ 	]*\([0-9]*\)[ 	].*/\1/p'`
	new_size=`wc -c $3 | sed -n 's/[ 	]*\([0-9]*\)[ 	].*/\1/p'`
	if [ ! -z "${new_size}" -a ! -z "${orig_size}" ] ; then
		new_size=`expr ${new_size} '*' 5 / 4`
		if [ "${new_size}" -le "${orig_size}" ] ; then
			echo
			echo "*********** WARNING *************"
			echo "****** initrd is smaller ********"
			echo "***** check configuration *******"
			echo "*********************************"
			echo
		fi
	fi

	# Check content to ensure that the module was updated
	temp=/tmp/`basename \`basename $3 .gz\` .img`.$$
	trap 'umount ${temp}.img ; rm -rf ${temp}.img ${temp}.dir ; exit' 1 2 3 15
	gunzip <$3 >${temp}.img 2>/dev/null
	mkdir ${temp}.dir 2>/dev/null
	mount -o loop ${temp}.img ${temp}.dir >/dev/null 2>/dev/null
	mod_dir=/lib/modules/$2/kernel/drivers/scsi
	if [ -d ${mod_dir}/${product} ] ; then
		mod_dir=${mod_dir}/${product}
	fi
	cmp "`find ${temp}.dir | grep ${product}.k*o | tail -1`" ${mod_dir}/${product}.[ko]* >/dev/null 2>/dev/null
	retval=$?
	umount ${temp}.img
	rm -rf ${temp}.img ${temp}.dir
	trap 'exit' 1 2 3 15
	if [ "${retval}" != "0" ] ; then
		echo
		echo "******* PROBLEM *********"
		echo "initrd ${product} module in $3 does not match package!"
		echo "*************************"
	fi

	return 0

}

run_lilo() {

	echo
        echo "Running lilo..."
	if [ "${root_dir}" != "" ] ; then
		lilo -r ${root_dir} -w
	else
		lilo
	fi
	if [ $? -ne 0 ] ; then
		echo Problem running lilo.  Quitting.
		exit 1
	fi
	echo
        echo "Reboot to load Ram Disk."
	echo
}


# parse the arguments
while [ $# -gt 0 ]; do
	case $1 in
		--version)
			echo `basename $0`": version ${VERSION}"
			exit 0
			;;

		--verbose)
			verbose=1
			;;

		--root*)
			if echo $1 | grep '=' >/dev/null ; then
				root_dir=`echo $1 | sed 's/^--root=//'`
			else
				root_dir=$2
				shift
			fi
			;;

		--list_modules)
			list_prebuilt_modules
			exit 0
			;;

		--noinitrd)
			do_mkinitrd=0
			do_lilo=0
			;;

		--nolilo)
			do_lilo=0
			;;

		--force)
			force=1
			;;

		--help)
			usage
			;;
		*)
			usage
			;;
	esac

	shift
done

if [ ${verbose} -eq 1 ] ; then
	echo "Options: do_mkinitrd="${do_mkinitrd}" do_lilo="${do_lilo}" root dir="${root_dir}
fi

# Check if and which bootloader we are using

loader_mode=-1
if [ -e ${lilo_conf} ] ; then
	lilo_boot_disk=`awk '\
	BEGIN { \
		FS="[ \11=]*"; \
		liloboot="?" ;\
	} \
	/boot[ \11]*=/ { \
		if ( $1 == "" ) { \
			liloboot=$3; \
		} else {\
			liloboot=$2 ;\
		} \
		print liloboot; \
	}' "/etc/lilo.conf";`
	dd bs=10 count=1 if=${lilo_boot_disk} 2>/dev/null | grep LILO >/dev/null 2>/dev/null
	loader_mode=$?
fi

if [ ${loader_mode} = 0 ] ; then
	echo "Using LILO configuration"
	awk_script="${installer_path}/lilo.awk"
	awk_file=${lilo_conf}
else
	do_lilo=0
	if [ -e ${grub_conf} ] ; then
		echo "Using GRUB configuration"
		use_grub=1
		loader_mode=1
		awk_script="${installer_path}/grub.awk"
		awk_file=${grub_conf}
	else
		awk_script=""
		awk_file=""
		loader_mode=-1
	fi
fi

modify_rc_config
modify_modules_conf
write_magic

file=file
[ -x /usr/bin/file ] && file=/usr/bin/file
[ -x ${root_dir}/usr/bin/file ] && file=${root_dir}/usr/bin/file

rm -f /tmp/${product}.done /tmp/${product}.fail >/dev/null 2>/dev/null

awk -f ${awk_script} ${awk_file} | sort +2 | while read image initrd label; do

	if [ ${image} = "?" ] ; then
		global_initrd=${initrd}
		continue
	fi
	echo
	file_magic=`${file} -L -m ${tmp_magic} ${image} 2>/dev/null`" \""
	kernel=`echo ${file_magic} | cut -d \" -f 2 `
	release=`echo ${file_magic} | cut -d \" -f 4 `

	echo "Configuration ["${label}"]"
	if [ -z "${kernel}" ] ; then
		echo
		echo ***WARNING: ${image} "is not a kernel. Skipping."
		echo
		# echo "${kernel}" >>/tmp/${product}.fail
		continue
	fi

	if [ ! -z "`echo ${label} | grep -i -e -Recover 2>/dev/null`" ] ; then
		echo
		echo ***NOTICE: ${image} "is Safe Recover selection. Skipping."
		echo
		echo "${kernel}" >>/tmp/${product}.done
		continue
	fi

	if [ ${verbose} -eq 1 ] ; then
		tmp=
		if [ ! -z "${release}" ] ; then
			tmp="(${release})"
		fi
		echo "label="${label}" image="${image}" version="\[${kernel}\]"${tmp} initrd="${initrd}
	fi

	if [ ${do_mkinitrd} -eq 1 ] ; then
		if [ ${initrd} = "?" ] ; then
			if [ -z ${global_initrd} ] ; then
				echo
				echo "****** PROBLEM *******"
				echo Kernel ${image} is not configured to use initrd.  Skipping.
				echo
				echo "TO FIX THIS PROBLEM:"
				echo "Modify ${awk_file} for this config to use an initrd"
				echo "(e.g. initrd=/boot/initrd-${kernel}.img)"
				echo "**********************"
				continue
			fi
			if [ -z ${last_initrd_kernel} ] ; then
				initrd=${global_initrd}
			elif [ ${last_initrd_kernel} = ${kernel} ] ; then
				initrd=${global_initrd}
			else
				echo "****** PROBLEM *******"
				echo A globally declared initrd is used for two different
				echo kernel versions.  Please fix this in ${awk_file}
				echo and rerun the install.
				echo "**********************"
				exit 1
			fi
			if [ ${verbose} -eq 1 ] ; then
				echo "Setting initrd to global initrd->"${initrd}
			fi
			last_initrd_kernel=${kernel}

		fi

		if [ x${initrd} = x${last_initrd} ] ; then
			if [ x${kernel} = x${last_kernel} ] ; then
				echo "Same as last initrd. Skipping."
				continue
			elif [ ${force} -ne 1 ] ; then
				echo
				echo "****** PROBLEM *******"
				echo In ${awk_file} there are two different kernel versions
				echo assigned to the same initrd - fix this and rerun installer.
				echo Or use the --force option
				echo "**********************"
				exit 1
			fi
		fi
	fi


	echo ${image}" -> "${kernel}
	last_kernel=${kernel}
	last_initrd=${initrd}
	tmpVersion=`echo ${kernel} | sed -e "s/^2.2//"`
	if [ ${tmpVersion} != ${kernel} ] ; then
		installPath=${root_dir}"/lib/modules/"${kernel}"/scsi"
	else
		installPath=${root_dir}"/lib/modules/"${kernel}"/kernel/drivers/scsi"
	fi

	machine=`uname -m`
	if [ ${machine} == "i686" ] ; then
		case "${release}" in
		*-athlon-*)	machine="athlon" ;;
		esac
		# athlon check
		grep "AMD Athlon" /proc/cpuinfo 2>/dev/null >/dev/null
		if [ $? -eq 0 ] ; then
			machine="athlon"
		fi
		grep "AMD-K7" /proc/cpuinfo 2>/dev/null >/dev/null
		if [ $? -eq 0 ] ; then
			machine="athlon"
		fi
	elif [ ${machine} == "x86_64" ] ; then
		# ia32e check
		grep "Xeon" /proc/cpuinfo 2>/dev/null >/dev/null
		if [ $? -eq 0 ] ; then
			machine="ia32e"
		fi
	fi

	install_module ${installPath} ${kernel} ${machine} ${release}
	if [ $? -ne 0 ] ; then
		echo "${kernel}" >>/tmp/${product}.fail
		continue
	fi
	echo "Installing in "${installPath}
#	depmod -b ${root_dir}/lib/modules/"${kernel}"

	echo "${kernel}" >>/tmp/${product}.done

	if [ ${do_mkinitrd} -eq 1 ] ; then
		run_mkinitrd ${image} ${kernel} ${initrd}
		if [ $? -ne 0 ] ; then
			exit 1
		fi
	fi

done

if [ $? -eq 1 ] ; then
	clean_magic
	exit 1
fi
clean_magic

echo
echo "Checking ${root_dir}/lib/modules for any remaining kernels not yet installed"

ls ${root_dir}/lib/modules | while read kernel ; do
	case ${kernel} in
	*[0-9].[0-9]*.[0-9]*)
		;;
	*)	continue ;;
	esac
	if [ ! -d "${root_dir}/lib/modules/${kernel}" ] ; then
		continue
	fi
	if [ ! -d "${root_dir}/lib/modules/${kernel}/kernel" ] ; then
		continue
	fi
	if [ -s /tmp/${product}.done ] ; then
		while read one ; do
			if [ "${one}" == "${kernel}" ] ; then
				break
			fi
		done < /tmp/${product}.done
		if [ "${one}" == "${kernel}" ] ; then
			continue
		fi
	fi

	echo
	echo "Configuration ["${kernel}"]"

	tmpVersion=`echo ${kernel} | sed -e "s/^2.2//"`
	if [ ${tmpVersion} != ${kernel} ] ; then
		installPath=${root_dir}"/lib/modules/"${kernel}"/scsi"
	else
		installPath=${root_dir}"/lib/modules/"${kernel}"/kernel/drivers/scsi"
	fi

	machine=`uname -m`
	if [ ${machine} == "i686" ] ; then
		# athlon check
		grep "AMD Athlon" /proc/cpuinfo 2>/dev/null >/dev/null
		if [ $? -eq 0 ] ; then
			machine="athlon"
		fi
		grep "AMD-K7" /proc/cpuinfo 2>/dev/null >/dev/null
		if [ $? -eq 0 ] ; then
			machine="athlon"
		fi
	elif [ ${machine} == "x86_64" ] ; then
		# ia32e check
		grep "Xeon" /proc/cpuinfo 2>/dev/null >/dev/null
		if [ $? -eq 0 ] ; then
			machine="ia32e"
		fi
	fi
	case ${kernel} in
	*-athlon)	machine="athlon" ;;
	*-x86_64)	machine="x86_64" ;;
	*-ia32e)	machine="ia32e" ;;
	*-ia64)		machine="ia64" ;;
	esac

	install_module ${installPath} ${kernel} ${machine}
	if [ $? -ne 0 ] ; then
		echo "${kernel}" >>/tmp/${product}.fail
		continue
	fi
	echo "Installing in "${installPath}

	echo "${kernel}" >>/tmp/${product}.done

done

module_count=`( wc -l /tmp/${product}.done 2>/dev/null || echo 0 /tmp/${product}.done ) | sed 's/[ 	]*\([0-9]*\)[ 	].*/\1/'`
failed_count=`( sort -u /tmp/${product}.fail 2>/dev/null | wc -l - 2>/dev/null ) | sed -n 's/[ 	]*\([0-9]*\)[ 	].*/\1/p'`
if [ -z "${failed_count}" ] ; then
	failed_count=0
fi
rm -f /tmp/${product}.done /tmp/${product}.fail >/dev/null 2>/dev/null
if [ ${module_count} -eq 0 ] ; then
	echo
	echo "***************************************"
	echo "******** NO MODULES INSTALLED *********"
	if [ ${failed_count} -eq 0 ] ; then
		echo "******** NO MODULES FAILED *********"
	elif [ ${failed_count} -eq 1 ] ; then
		echo "********  A MODULE FAILED  *********"
	else
		echo "********  ${failed_count} MODULES FAILED *********"
	fi
	echo "***************************************"
	echo
	echo "Done."
	exit 1
elif [ ${failed_count} -ne 0 ] ; then
	echo
	echo "***************************************"
	if [ ${failed_count} -eq 1 ] ; then
		echo "********  A MODULE FAILED  *********"
	else
		echo "********  ${failed_count} MODULES FAILED *********"
	fi
	echo "***************************************"
	echo
fi


echo

if [ ${do_lilo} -eq 1 ] ; then
	run_lilo
fi


#./create_device_nodes

echo Done.
exit
